# !/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
qg_botsdk Library
~~~~~~~~~~~~~~~~~~~~~

一款简洁、容易上手，适用于QQ官方频道机器人的Python轻量级SDK；只需导入qg_botsdk.qg_bot.BOT即可快速使用

.. seealso::
    更多教程和相关资讯可参阅：
    https://qg-botsdk.readthedocs.io/zh_CN/latest/index.html

:copyright: (c) 2022 by GLGDLY.
:license: MIT, see LICENSE for more details.
"""

from .version import __version__
from .qg_bot import BOT
from .model import Model
from .logger import Logger

__title__ = "requests"
__description__ = "easy-to-use SDK for Tencent QQ guild robot"
__url__ = "https://github.com/GLGDLY/qg_botsdk"
__author__ = "GDLY"
__author_email__ = "tzlgdly@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2022 GLGDLY"


name = "qg_botsdk"
__all__ = ("BOT", "Model", "Logger")



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
from sys import exc_info
from traceback import extract_tb
from inspect import stack
from json import dumps
from json.decoder import JSONDecodeError
from functools import wraps
from re import split as re_split


def __getattr__(identifier: str) -> object:
    if re_split(r'[/\\]', stack()[1].filename)[-1] not in ('qg_bot.py', 'qg_bot_ws.py', 'api.py', 'async_api.py',
                                                           '<frozen importlib._bootstrap>'):
        raise AssertionError("此为SDK内部使用文件，无法使用")

    return globals()[identifier.__path__]


def template_wrapper(func):
    @wraps(func)
    def wrap(*args):
        try:
            return func(*args)
        except (JSONDecodeError, AttributeError, KeyError):
            return_ = args[0]
            code = return_.status_code if hasattr(return_, 'status_code') else getattr(return_, 'status', None)
            trace_id = return_.headers.get('X-Tps-Trace-Id', None) if hasattr(return_, 'headers') else None
            return objectize({'data': None, 'trace_id': trace_id, 'http_code': code, 'result': False})

    return wrap


def security_wrapper(func):
    @wraps(func)
    def wrap(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (JSONDecodeError, KeyError, AttributeError) as e:
            name = func.__name__
            logger = getattr(args[0], 'logger', None)
            if name == '__security_check_code':
                if logger:
                    logger.error('无法调用内容安全检测接口（备注：请检查机器人密钥是否正确）')
            else:
                if logger:
                    logger.error(f'调用内容安全检测接口失败，详情：{exception_handler(e)}')
                return False
    return wrap


def exception_handler(error):
    error_info = extract_tb(exc_info()[-1])[-1]
    return "[error:{}] File \"{}\", line {}, in {}".format(error.__repr__(), *error_info[:3])


def exception_processor(func):
    @wraps(func)
    def wrap(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger = getattr(args[0], 'logger', None)
            if logger:
                logger.error(e.__repr__())
                logger.debug(exception_handler(e))

    return wrap


class object_class(type):
    def __repr__(self):
        return self.__doc__


def objectize(data: dict):
    doc = dumps(data)
    if isinstance(data, dict):
        for keys, values in data.items():
            if keys.isnumeric():
                return data
            if isinstance(values, dict):
                data[keys] = objectize(values)
            elif isinstance(values, list):
                for i, items in enumerate(values):
                    if isinstance(items, dict):
                        data[keys][i] = objectize(items)
        data['__doc__'] = doc
        object_data = object_class('object', (object,), data)
        return object_data
    else:
        return None


def treat_msg(raw_msg: str):
    if not raw_msg:
        return ''
    if raw_msg[0] == '/':
        raw_msg = raw_msg[1:]
    return raw_msg.replace('&amp;', '&').replace('&lt;', '<').replace('&gt;', '>').replace('\xa0', ' ').strip()


@template_wrapper
def http_temp(return_, code: int):
    trace_id = return_.headers.get("X-Tps-Trace-Id", None)
    real_code = return_.status_code
    if real_code == code:
        return objectize({'data': None, 'trace_id': trace_id, 'http_code': real_code, 'result': True})
    else:
        return_dict = return_.json()
        return objectize({'data': return_dict, 'trace_id': trace_id, 'http_code': real_code, 'result': False})


@template_wrapper
async def async_http_temp(return_, code: int):
    trace_id = return_.headers.get("X-Tps-Trace-Id", None)
    real_code = return_.status
    if real_code == code:
        return objectize({'data': None, 'trace_id': trace_id, 'http_code': real_code, 'result': True})
    else:
        return_dict = await return_.json()
        return objectize({'data': return_dict, 'trace_id': trace_id, 'http_code': real_code, 'result': False})


@template_wrapper
def regular_temp(return_):
    trace_id = return_.headers.get("X-Tps-Trace-Id", None)
    return_dict = return_.json()
    if isinstance(return_dict, dict) and 'code' in return_dict:
        result = False
    else:
        result = True
    return objectize({'data': return_dict, 'trace_id': trace_id, 'http_code': return_.status_code, 'result': result})


@template_wrapper
async def async_regular_temp(return_):
    trace_id = return_.headers.get("X-Tps-Trace-Id", None)
    return_dict = await return_.json()
    if isinstance(return_dict, dict) and 'code' in return_dict:
        result = False
    else:
        result = True
    return objectize({'data': return_dict, 'trace_id': trace_id, 'http_code': return_.status, 'result': result})


@template_wrapper
def empty_temp(return_):
    trace_id = return_.headers.get("X-Tps-Trace-Id", None)
    return_dict = return_.json()
    if not return_dict:
        result = True
        return_dict = None
    else:
        result = False
    return objectize({'data': return_dict, 'trace_id': trace_id, 'http_code': return_.status_code, 'result': result})


@template_wrapper
async def async_empty_temp(return_):
    trace_id = return_.headers.get("X-Tps-Trace-Id", None)
    return_dict = await return_.json()
    if not return_dict:
        result = True
        return_dict = None
    else:
        result = False
    return objectize({'data': return_dict, 'trace_id': trace_id, 'http_code': return_.status, 'result': result})


def sdk_error_temp(message):
    return objectize({'data': {'code': -1, 'message': f'这是来自SDK的错误信息：{message}'}, 'trace_id': None,
                      'http_code': None, 'result': False})



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
from logging import getLogger, Formatter, StreamHandler, FileHandler
from os import makedirs, PathLike, getcwd, sep
from os.path import exists, join, isdir
from time import strftime, localtime
from functools import wraps
from colorama import init as color_init
from typing import Union, List
from re import split as re_split

color_init(strip=False)


def _log_wrapper(func):
    @wraps(func)
    def wrap(self, msg):
        str_time = strftime('%m-%d', localtime())
        if str_time != self._previous_time:
            self._logger.removeHandler(self._logh)
            self._new_logh(str_time)
        try:
            func(self, msg)
        except UnicodeDecodeError:
            pass

    return wrap


class Logger:
    def __init__(self, bot_app_id: str, file_path: Union[str, PathLike] = None, disable_logger: List[str] = None):
        """
        用作logging输出的类，支持不同level的颜色log输出，可自定义格式

        :param bot_app_id: 机器人ID
        :param file_path: 可选填，自定义log文件输出路径，不填默认为：当前目录/log/[bot_app_id]/
        :param disable_logger: 可选填，需要disable的logger名称列表，不填默认无
        """
        getLogger("asyncio").disabled = True
        if disable_logger:
            for items in disable_logger:
                getLogger(items).disabled = True
        self._bot_app_id = bot_app_id
        self._logger = getLogger(__name__)
        self._logger.setLevel('DEBUG')
        self._format = '[%(asctime)s] [%(levelname)s] %(message)s'
        self._date_format = '%m-%d %H:%M:%S'
        self._cmdh = StreamHandler()
        self._cmdh.setFormatter(self._Stream_Formatter())
        self._cmdh.setLevel('INFO')
        self.file_path = file_path
        if file_path is not None:
            file_path = re_split(r'[\\|/]', file_path)
            if file_path[0] == '.':
                self.file_path = join(*file_path)
            else:
                file_path[0] += sep
                self.file_path = join(sep, *file_path)
        else:
            self.file_path = join(getcwd(), 'log', self._bot_app_id)
        if not exists(self.file_path):
            makedirs(self.file_path)
        assert isdir(self.file_path), '自定义Log输出路径必须为一个directory资料夹'
        self._logh = None
        self._new_logh(strftime('%m-%d', localtime()))
        self._logger.addHandler(self._cmdh)
        self._previous_time = strftime('%m-%d', localtime())

    class _Stream_Formatter(Formatter):
        FORMATS = {20: '\033[1;32m[%(asctime)s] [%(levelname)s]\033[0m %(message)s',
                   30: '\033[1;33m[%(asctime)s] [%(levelname)s]\033[0m %(message)s',
                   40: '\033[1;31m[%(asctime)s] [%(levelname)s]\033[0m %(message)s'}

        def __init__(self, formats: dict = None, date_format: str = None):
            super().__init__()
            if formats is not None:
                self.FORMATS[20] = formats.get(20, None) or self.FORMATS[20]
                self.FORMATS[30] = formats.get(30, None) or self.FORMATS[30]
                self.FORMATS[40] = formats.get(40, None) or self.FORMATS[40]
            self._date_format = date_format or '%m-%d %H:%M:%S'

        def format(self, record) -> str:
            log_fmt = self.FORMATS.get(record.levelno)
            formatter = Formatter(log_fmt, self._date_format)
            return formatter.format(record)

    def set_formatter(self, debug_format: str = None, info_format: str = None, warning_format: str = None,
                      error_format: str = None, date_format: str = None):
        """
        用于更改logger日志的输出格式

        - 更多关于格式的相关信息可参阅：
        - https://docs.python.org/zh-cn/3/library/logging.html#logging.Formatter
        - https://docs.python.org/zh-cn/3.6/library/logging.html#logrecord-attributes

        :param debug_format: debug层级的格式，此格式会输出到保存的日志文件中
        :param info_format: info层级StreamHandler的格式，此格式会输出到python console中显示
        :param warning_format: warning层级StreamHandler的格式，此格式会输出到python console中显示
        :param error_format: error层级StreamHandler的格式，此格式会输出到python console中显示
        :param date_format: 全局所有层级的日期格式
        """
        self._date_format = date_format or self._date_format
        self._format = debug_format or self._format
        formats = {20: info_format, 30: warning_format, 40: error_format}
        self._cmdh.setFormatter(self._Stream_Formatter(formats, date_format))
        self._logh.setFormatter(Formatter(self._format, self._date_format))

    def _new_logh(self, str_time):
        self._logh = FileHandler(join(self.file_path, f'{str_time}.log'), encoding='utf-8')
        self._logh.setFormatter(Formatter(self._format, self._date_format))
        self._logger.addHandler(self._logh)

    @_log_wrapper
    def debug(self, msg):
        self._logger.debug(msg)

    @_log_wrapper
    def info(self, msg):
        self._logger.info(msg)

    @_log_wrapper
    def warning(self, msg):
        self._logger.warning(msg)

    @_log_wrapper
    def error(self, msg):
        self._logger.error(msg)



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Dict, Any, Union, Tuple
try:
    from yaml import safe_load
except (ImportError, ModuleNotFoundError):
    safe_load = None


def read_yaml(yaml_path) -> Dict[str, Any]:
    assert safe_load, '如需使用read_yaml函数，必须先pip install pyyaml'
    with open(yaml_path, "r", encoding="utf-8") as f:
        return safe_load(f)


def convert_color(color: Union[Tuple[int, int, int], str]):
    colors = []
    if isinstance(color, tuple):
        if len(color) == 3:
            for items in color:
                if not isinstance(items, int) or items not in range(256):
                    raise TypeError('RGB颜色应为一个三位数的tuple，且当中每个数值都应该介乎于0和255之间，如(255,255,255)')
                colors.append(items)
        else:
            raise TypeError('RGB颜色应为一个三位数的tuple，且当中每个数值都应该介乎于0和255之间，如(255,255,255)')
    elif isinstance(color, str):
        colour = color.replace('#', '')
        if len(colour) == 6:
            for items in (colour[:2], colour[2:4], colour[4:]):
                try:
                    items = int(items, 16)
                except ValueError:
                    raise TypeError('该HEX颜色不存在，请检查其颜色值是否准确') from None
                if items not in range(256):
                    raise TypeError('该HEX颜色不存在，请检查其颜色值是否准确')
                colors.append(items)
        else:
            raise TypeError('HEX颜色应为一个 #加六位数字或字母 的string，如"#ffffff"')
    else:
        raise TypeError('颜色值应为RGB的三位数tuple，如(255,255,255)；或HEX的sting颜色，如"#ffffff"')
    return colors[0] + 256 * colors[1] + (256**2) * colors[2]



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
from os import getpid
from asyncio import get_event_loop, sleep
from time import sleep as t_sleep
from ssl import SSLContext
from requests import Session
from requests.adapters import HTTPAdapter, Retry
from typing import Any, Callable, Union, Optional
from . import _api_model
from .version import __version__
from .logger import Logger
from .model import Model
from .qg_bot_ws import BotWs
from ._utils import exception_processor
from .api import API

retry = Retry(total=4, connect=2, read=2)
adapter = HTTPAdapter(max_retries=retry)
pid = getpid()
print(f'本次程序进程ID：{pid} | SDK版本：{__version__} | 即将开始运行机器人……')
t_sleep(0.5)


class BOT:

    def __init__(self, bot_id: str, bot_token: str, bot_secret: Optional[str] = None, is_private: bool = False,
                 is_sandbox: bool = False, no_permission_warning: bool = True, is_async: bool = False,
                 is_retry: bool = True, is_log_error: bool = True, shard_no: int = 0, total_shard: int = 1,
                 max_workers: Optional[int] = None):
        """
        机器人主体，输入BotAppID和密钥，并绑定函数后即可快速使用

        :param bot_id: 机器人平台后台BotAppID（开发者ID）项，必填
        :param bot_token: 机器人平台后台机器人令牌项，必填
        :param bot_secret: 机器人平台后台机器人密钥项，如需要使用安全检测功能需填写此项
        :param is_private: 机器人是否为私域机器人，默认False
        :param is_sandbox: 是否开启沙箱环境，默认False
        :param no_permission_warning: 是否开启当机器人获取疑似权限不足的事件时的警告提示，默认开启
        :param is_async: 使用同步api还是异步api，默认False（使用同步）
        :param is_retry: 使用api时，如遇可重试的错误码是否自动进行重试，默认开启
        :param is_log_error: 使用api时，如返回的结果为不成功，可自动log输出报错信息，默认开启
        :param shard_no: 当前分片数，如不熟悉相关配置请不要轻易改动此项，默认0
        :param total_shard: 最大分片数，如不熟悉相关配置请不要轻易改动此项，默认1
        :param max_workers: 在同步模式下，允许同时运行的最大线程数
        """
        self.logger = Logger(bot_id)
        self.bot_id = bot_id
        self.bot_token = bot_token
        self.bot_secret = bot_secret
        self.is_private = is_private
        self.bot_url = r'https://sandbox.api.sgroup.qq.com' if is_sandbox else r'https://api.sgroup.qq.com'
        if not self.bot_id or not self.bot_token:
            raise type('IdTokenMissing', (Exception,), {})(
                '你还没有输入 bot_id 和 bot_token，无法连接使用机器人\n如尚未有相关票据，'
                '请参阅 https://qg-botsdk.readthedocs.io/zh_CN/latest/quick_start 了解相关详情')
        self.intents = 0
        self.shard_no = shard_no
        self.total_shard = total_shard
        self.__on_delete_function = None
        self.__on_msg_function = None
        self.__on_dm_function = None
        self.__on_guild_event_function = None
        self.__on_channel_event_function = None
        self.__on_guild_member_function = None
        self.__on_reaction_function = None
        self.__on_interaction_function = None
        self.__on_audit_function = None
        self.__on_forum_function = None
        self.__on_audio_function = None
        self.__repeat_function = None
        self.__on_start_function = None
        self.is_filter_self = True
        self.check_interval = 10
        self.running = False
        self.bot_headers = {'Authorization': f"Bot {self.bot_id}.{self.bot_token}"}
        self.__ssl = SSLContext()
        self.__session = Session()
        self.__session.headers = self.bot_headers
        self.__session.keep_alive = False
        self.__session.mount('http://', adapter)
        self.__session.mount('https://', adapter)
        self.__bot_class = None
        self.__loop = get_event_loop()
        self.msg_treat = True
        self.dm_treat = False
        self.no_permission_warning = no_permission_warning
        self.max_workers = max_workers
        self.is_async = is_async
        if not is_async:
            self.api: API = API(self.bot_url, bot_id, bot_secret, self.__session, self.logger, self._check_warning,
                                is_retry, is_log_error)
        else:
            from .async_api import AsyncAPI
            self.api: Union[AsyncAPI, API] = AsyncAPI(self.bot_url, bot_id, bot_secret, self.__ssl, self.bot_headers,
                                                      self.logger, self.__loop, self._check_warning,
                                                      is_retry, is_log_error)

    def __repr__(self):
        return f"<qg_botsdk.BOT object [id: {self.bot_id}, token: {self.bot_token}]>"

    @property
    def robot(self) -> _api_model.robot():
        return self.__bot_class.robot

    @exception_processor
    def __time_event_run(self):
        if self.is_async:
            self.__loop.create_task(self.__repeat_function())
        else:
            self.__repeat_function()

    async def __time_event_check(self):
        while self.running:
            self.__time_event_run()
            await sleep(self.check_interval)

    def bind_msg(self, on_msg_function: Callable[[Model.MESSAGE], Any], treated_data: bool = True,
                 all_msg: bool = None):
        """
        用作绑定接收消息的函数，将根据机器人是否公域自动判断接收艾特或所有消息

        :param on_msg_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        :param treated_data: 是否返回经转义处理的文本，如是则会在返回的Object中添加一个treated_msg的子类，默认True
        :param all_msg: 是否无视公私域限制，强制开启全部消息接收，默认None（不判断此项参数）
        """
        self.__on_msg_function = on_msg_function
        if not treated_data:
            self.msg_treat = False
        if all_msg is None:
            if not self.is_private:
                self.intents = self.intents | 1 << 30
                self.logger.info('消息（所有消息）接收函数订阅成功')
            else:
                self.intents = self.intents | 1 << 9
                self.logger.info('消息（艾特消息）接收函数订阅成功')
        else:
            self.intents = self.intents | 1 << 30
            self.logger.info('消息（所有消息）接收函数订阅成功')

    def bind_dm(self, on_dm_function: Callable[[Model.DIRECT_MESSAGE], Any], treated_data: bool = True):
        """
        用作绑定接收私信消息的函数

        :param on_dm_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        :param treated_data: 是否返回经转义处理的文本，如是则会在返回的Object中添加一个treated_msg的子类，默认True
        """
        self.__on_dm_function = on_dm_function
        self.intents = self.intents | 1 << 12
        if treated_data:
            self.dm_treat = True
        self.logger.info('私信接收函数订阅成功')

    def bind_msg_delete(self, on_delete_function: Callable[[Model.MESSAGE_DELETE], Any], is_filter_self: bool = True):
        """
        用作绑定接收消息撤回事件的函数，注册时将自动根据公域私域注册艾特或全部消息，但不会主动注册私信事件

        :param on_delete_function:类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        :param is_filter_self: 是否过滤用户自行撤回的消息，只接受管理撤回事件
        """
        self.__on_delete_function = on_delete_function
        self.is_filter_self = is_filter_self
        if self.is_private:
            self.intents = self.intents | 1 << 30
        else:
            self.intents = self.intents | 1 << 9
        self.logger.info('撤回事件订阅成功')

    def bind_guild_event(self, on_guild_event_function: Callable[[Model.GUILDS], Any]):
        """
        用作绑定接收频道信息的函数

        :param on_guild_event_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        """
        self.__on_guild_event_function = on_guild_event_function
        self.intents = self.intents | 1 << 0
        self.logger.info('频道事件订阅成功')

    def bind_channel_event(self, on_channel_event_function: Callable[[Model.CHANNELS], Any]):
        """
        用作绑定接收子频道信息的函数

        :param on_channel_event_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        """
        self.__on_channel_event_function = on_channel_event_function
        self.intents = self.intents | 1 << 0
        self.logger.info('子频道事件订阅成功')

    def bind_guild_member(self, on_guild_member_function: Callable[[Model.GUILD_MEMBERS], Any]):
        """
        用作绑定接收频道成员信息的函数

        :param on_guild_member_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        """
        self.__on_guild_member_function = on_guild_member_function
        self.intents = self.intents | 1 << 1
        self.logger.info('频道成员事件订阅成功')

    def bind_reaction(self, on_reaction_function: Callable[[Model.REACTION], Any]):
        """
        用作绑定接收表情表态信息的函数

        :param on_reaction_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        """
        self.__on_reaction_function = on_reaction_function
        self.intents = self.intents | 1 << 10
        self.logger.info('表情表态事件订阅成功')

    def bind_interaction(self, on_interaction_function: Callable[[Any], Any]):
        """
        用作绑定接收互动事件的函数，当前未有录入数据结构

        :param on_interaction_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        """
        self.__on_interaction_function = on_interaction_function
        self.intents = self.intents | 1 << 26
        self.logger.info('互动事件订阅成功')

    def bind_audit(self, on_audit_function: Callable[[Model.MESSAGE_AUDIT], Any]):
        """
        用作绑定接收审核事件的函数

        :param on_audit_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        """
        self.__on_audit_function = on_audit_function
        self.intents = self.intents | 1 << 27
        self.logger.info('审核事件订阅成功')

    def bind_forum(self, on_forum_function: Callable[[Model.FORUMS_EVENT], Any]):
        """
        用作绑定接收论坛事件的函数，一般仅私域机器人能注册此事件

        .. note::
            当前仅可以接收FORUM_THREAD_CREATE、FORUM_THREAD_UPDATE、FORUM_THREAD_DELETE三个事件

        :param on_forum_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        """

        self.__on_forum_function = on_forum_function
        self.intents = self.intents | 1 << 28
        self.logger.info('论坛事件订阅成功')
        if not self.is_private and self.no_permission_warning:
            self.logger.warning('请注意，一般公域机器人并不能注册论坛事件，请检查自身是否拥有相关权限')

    def bind_audio(self, on_audio_function: Callable[[Model.AUDIO_ACTION], Any]):
        """
        用作绑定接收论坛事件的函数

        :param on_audio_function: 类型为function，该函数应包含一个参数以接收Object消息数据进行处理
        """
        self.__on_audio_function = on_audio_function
        self.intents = self.intents | 1 << 29
        self.logger.info('音频事件订阅成功')
        if self.no_permission_warning:
            self.logger.warning('请注意，一般机器人并不能注册音频事件（需先进行申请），请检查自身是否拥有相关权限（如已申请可忽略此消息）')

    def register_repeat_event(self, time_function: Callable[[], Any], check_interval: Union[float, int] = 10):
        """
        用作注册重复事件的函数，注册并开始机器人后，会根据间隔时间不断调用注册的函数

        :param time_function: 类型为function，该函数不应包含任何参数
        :param check_interval: 每多少秒检查调用一次时间事件函数，默认10
        """
        self.__repeat_function = time_function
        self.check_interval = check_interval
        self.logger.info('重复运行事件注册成功')

    def register_start_event(self, on_start_function: Callable[[], Any]):
        """
        用作注册机器人开始时运行的函数，此函数不应有无限重复的内容

        :param on_start_function: 类型为function，该函数不应包含任何参数
        """
        self.__on_start_function = on_start_function
        self.logger.info('初始事件注册成功')

    def _check_warning(self, name: str):
        if not self.is_private and self.no_permission_warning:
            self.logger.warning(f'请注意，一般公域机器人并不能使用{name}API，请检查自身是否拥有相关权限')

    def start(self):
        """
        开始运行机器人的函数，在唤起此函数后的代码将不能运行，如需运行后续代码，请以多进程方式唤起此函数，以下是一个简单的唤起流程：

        >>> from qg_botsdk import BOT
        >>> bot = BOT(bot_id='xxx', bot_token='xxx')
        >>> bot.start()

        .. seealso::
            更多教程和相关资讯可参阅：
            https://qg-botsdk.readthedocs.io/zh_CN/latest/index.html
        """
        try:
            if not self.running and not self.__bot_class:
                self.running = True
                gateway = self.__session.get(self.bot_url + '/gateway/bot').json()
                if 'url' not in gateway:
                    raise type('IdTokenError', (Exception,), {})(
                        '你输入的 bot_id 和/或 bot_token 错误，无法连接使用机器人\n如尚未有相关票据，'
                        '请参阅 https://qg-botsdk.readthedocs.io/zh_CN/latest/quick_start 了解相关详情')
                url = gateway["url"]
                self.logger.debug('[机器人ws地址] ' + url)
                if self.__repeat_function is not None:
                    self.__loop.create_task(self.__time_event_check())
                self.__bot_class = BotWs(self.__session, self.__ssl, self.logger, self.total_shard, self.shard_no, url,
                                         self.bot_id, self.bot_token, self.bot_url, self.__on_msg_function,
                                         self.__on_dm_function, self.__on_delete_function, self.is_filter_self,
                                         self.__on_guild_event_function, self.__on_channel_event_function,
                                         self.__on_guild_member_function, self.__on_reaction_function,
                                         self.__on_interaction_function, self.__on_audit_function,
                                         self.__on_forum_function, self.__on_audio_function, self.intents,
                                         self.msg_treat, self.dm_treat, self.__on_start_function, self.is_async,
                                         self.max_workers if self.max_workers else 10)
                self.__bot_class.starter()
            else:
                self.logger.error('当前机器人已在运行中！')
        except KeyboardInterrupt:
            exit()

    def close(self):
        """
        结束运行机器人的函数

        .. seealso::
            更多教程和相关资讯可参阅：
            https://qg-botsdk.readthedocs.io/zh_CN/latest/index.html
        """
        if self.running:
            self.running = False
            self.logger.info('WS链接已开始结束进程，请等待另一端完成握手并等待 TCP 连接终止')
            self.__bot_class.running = False
            timeout = self.__loop.time() + 60
            while self.__loop.time() > timeout:
                t_sleep(1)
                if not self.__bot_class or self.__bot_class.ws.closed:
                    self.logger.info('WS链接已结束')
                    return
            self.__bot_class = None
            self.logger.info('判断超时，WS链接已强制结束')
        else:
            self.logger.error('当前机器人没有运行！')



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
from inspect import stack
from typing import List, Optional
from re import split as re_split

apis = {('获取用户ID', 'get_bot_id'): [False, '此API不需要请求权限'],
        ('获取用户信息', 'get_bot_info'): ['GET', '/users/@me'],
        ('获取用户频道列表', 'get_bot_guilds'): ['GET', '/users/@me/guilds'],
        ('获取频道详情', 'get_guild_info'): ['GET', '/guilds/{guild_id}'],
        ('获取子频道列表', 'get_guild_channels'): ['GET', '/guilds/{guild_id}/channels'],
        ('获取子频道详情', 'get_channels_info'): ['GET', '/channels/{channel_id}'],
        ('创建子频道', 'create_channels'): ['POST', '/guilds/{guild_id}/channels'],
        ('修改子频道', 'patch_channels'): ['PATCH', '/channels/{channel_id}'],
        ('删除子频道', 'delete_channels'): ['DELETE', '/channels/{channel_id}'],
        ('获取频道成员列表', 'get_guild_members'): ['GET', '/guilds/{guild_id}/members'],
        ('获取频道成员详情', 'get_member_info'): ['GET', '/guilds/{guild_id}/members/{user_id}'],
        ('删除频道成员', 'delete_member'): ['DELETE', '/guilds/{guild_id}/members/{user_id}'],
        ('获取频道身份组列表', 'get_guild_roles'): ['GET', '/guilds/{guild_id}/roles'],
        ('创建频道身份组', 'create_role'): ['POST', '/guilds/{guild_id}/roles'],
        ('修改频道身份组', 'patch_role'): ['PATCH', '/guilds/{guild_id}/roles/{role_id}'],
        ('删除频道身份组', 'delete_role'): ['DELETE', '/guilds/{guild_id}/roles/{role_id}'],
        ('创建频道身份组成员', 'create_role_member'): ['PUT', '/guilds/{guild_id}/members/{user_id}/roles/{role_id}'],
        ('删除频道身份组成员', 'delete_role_member'): ['DELETE', '/guilds/{guild_id}/members/{user_id}/roles/{role_id}'],
        ('获取子频道用户权限', 'get_channel_member_permission'):
            ['GET', '/channels/{channel_id}/members/{user_id}/permissions'],
        ('修改子频道用户权限', 'put_channel_member_permission'):
            ['PUT', '/channels/{channel_id}/members/{user_id}/permissions'],
        ('获取子频道身份组权限', 'get_channel_role_permission'):
            ['GET', '/channels/{channel_id}/roles/{role_id}/permissions'],
        ('修改子频道身份组权限', 'put_channel_role_permission'):
            ['PUT', '/channels/{channel_id}/roles/{role_id}/permissions'],
        ('获取指定消息', 'get_message_info'): ['GET', '/channels/{channel_id}/messages/{message_id}'],
        ('发送普通消息', 'send_msg', '发送embed模板消息', 'send_embed', '发送 23 链接+文本列表模板ark消息', 'send_ark_23',
         '发送 24 文本+缩略图模板ark消息', 'send_ark_24', '发送 37 大图模板ark消息', 'send_ark_37'):
            ['POST', '/channels/{channel_id}/messages'],
        ('撤回消息', 'delete_msg'): ['DELETE', '/channels/{channel_id}/messages/{message_id}'],
        ('获取频道消息频率设置', 'get_guild_setting'): ['GET', '/guilds/{guild_id}/message/setting'],
        ('创建私信会话', 'create_dm_guild'): ['POST', '/users/@me/dms'],
        ('发送私信消息', 'send_dm'): ['POST', '/dms/{guild_id}/messages'],
        ('撤回私信消息', 'delete_dm_msg'): ['DELETE', '/dms/{guild_id}/messages/{message_id}'],
        ('禁言全员', 'mute_all_member'): ['PATCH', '/guilds/{guild_id}/mute'],
        ('禁言指定成员', 'mute_member'): ['PATCH', '/guilds/{guild_id}/members/{user_id}/mute'],
        ('禁言批量成员', 'mute_members'): ['PATCH', '/guilds/{guild_id}/mute'],
        ('创建频道公告', 'create_announce'): ['POST', '/guilds/{guild_id}/announces'],
        ('删除频道公告', 'delete_announce'): ['DELETE', '/guilds/{guild_id}/announces/{message_id}'],
        ('添加精华消息', 'create_pinmsg'): ['PUT', '/channels/{channel_id}/pins/{message_id}'],
        ('删除精华消息', 'delete_pinmsg'): ['DELETE', '/channels/{channel_id}/pins/{message_id}'],
        ('获取精华消息', 'get_pinmsg'): ['GET', '/channels/{channel_id}/pins'],
        ('获取频道日程列表', 'get_schedules'): ['GET', '/channels/{channel_id}/schedules'],
        ('获取日程详情', 'get_schedule_info'): ['GET', '/channels/{channel_id}/schedules/{schedule_id}'],
        ('创建日程', 'create_schedule'): ['POST', '/channels/{channel_id}/schedules'],
        ('修改日程', 'patch_schedule'): ['PATCH', '/channels/{channel_id}/schedules/{schedule_id}'],
        ('删除日程', 'delete_schedule'): ['DELETE', '/channels/{channel_id}/schedules/{schedule_id}'],
        ('发表表情表态', 'create_reaction'):
            ['PUT', '/channels/{channel_id}/messages/{message_id}/reactions/{type}/{id}'],
        ('删除表情表态', 'delete_reaction'):
            ['DELETE', '/channels/{channel_id}/messages/{message_id}/reactions/{type}/{id}'],
        ('拉取表情表态用户列表', 'get_reaction_users'):
            ['GET', '/channels/{channel_id}/messages/{message_id}/reactions/{type}/{id}'],
        ('音频控制', 'control_audio'): ['POST', '/channels/{channel_id}/audio'],
        ('机器人上麦', 'bot_on_mic'): ['PUT', '/channels/{channel_id}/mic'],
        ('机器人下麦', 'bot_off_mic'): ['DELETE', '/channels/{channel_id}/mic'],
        ('获取帖子列表', 'get_threads'): ['GET', '/channels/{channel_id}/threads'],
        ('获取帖子详情', 'get_thread_info'): ['GET', '/channels/{channel_id}/threads/{thread_id}'],
        ('发表帖子', 'create_thread'): ['PUT', '/channels/{channel_id}/threads'],
        ('删除帖子', 'delete_thread'): ['DELETE', '/channels/{channel_id}/threads/{thread_id}'],
        ('获取频道可用权限列表', 'get_guild_permissions', '创建频道 API 接口权限授权链接', 'create_permission_demand'):
            [False, '此API不需要请求权限']}


def __getattr__(identifier: str) -> object:
    if re_split(r'[/\\]', stack()[1].filename)[-1] not in ('qg_bot.py', 'api.py', 'async_api.py',
                                                           '<frozen importlib._bootstrap>'):
        raise AssertionError("此为SDK内部使用文件，无法使用，使用机器人Model库请from model import Model")

    return globals()[identifier.__path__]


def api_converter(api: str):
    if api[-2:] in ['()', '（）']:
        api = api[:-2]
    for keys, values in apis.items():
        if api in keys:
            return values[0], values[1]
    return False, '该API并不存在'


def api_converter_re(method: str, path: str):
    for keys, values in apis.items():
        if [method, path] == values:
            return keys[1]


class _EmptyReturnTemplate:
    code: int
    message: str


def robot():
    class Robot:
        id: str
        username: str
        avatar: str

    return Robot


def get_bot_info():
    class GetBotInfo:
        class data:
            id: str
            username: str
            avatar: str
            union_openid: str
            union_user_account: str
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return GetBotInfo


def get_bot_guilds():
    class GetBotGuilds:
        class __guild:
            id: str
            name: str
            icon: str
            owner_id: str
            owner: bool
            joined_at: str
            member_count: str
            max_members: str
            description: str
            code: int
            message: str

        data: List[__guild]
        http_code: List[int]
        trace_id: List[str]
        result: List[bool]

    return GetBotGuilds


def get_guild_info():
    class GetGuildInfo:
        class data:
            id: str
            name: str
            icon: str
            owner_id: str
            owner: bool
            joined_at: str
            member_count: int
            max_members: int
            description: str
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return GetGuildInfo


def get_guild_channels():
    class GetGuildChannels:
        class __channels:
            id: str
            guild_id: str
            name: str
            type: int
            position: int
            parent_id: str
            owner_id: str
            sub_type: int
            code: int
            message: str

        data: List[__channels]
        http_code: int
        trace_id: str
        result: bool

    return GetGuildChannels


class _Channels:
    id: str
    guild_id: str
    name: str
    type: int
    position: int
    parent_id: str
    owner_id: str
    op_user_id: Optional[str]
    sub_type: int
    private_type: int
    private_user_ids: List[str]
    speak_permission: int
    application_id: str
    permissions: str
    code: int
    message: str


def get_channels_info():
    channels = _Channels

    class GetChannelsInfo:
        data: channels
        http_code: int
        trace_id: str
        result: bool

    return GetChannelsInfo


def create_channels():
    channels = _Channels

    class CreateChannels:
        data: channels
        http_code: int
        trace_id: str
        result: bool

    return CreateChannels


def patch_channels():
    channels = _Channels

    class PatchChannels:
        data: channels
        http_code: int
        trace_id: str
        result: bool

    return PatchChannels


def delete_channels():
    temp = _EmptyReturnTemplate

    class DeleteChannels:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return DeleteChannels


class _Member:
    class user:
        id: str
        username: str
        avatar: str
        bot: bool

    nick: str
    roles: List[str]
    joined_at: str
    deaf: bool
    mute: bool
    pending: bool
    code: int
    message: str


def get_guild_members():
    member = _Member

    class GetGuildMembers:
        data: List[member]
        http_code: List[int]
        trace_id: List[str]
        result: List[bool]

    return GetGuildMembers


def get_member_info():
    member = _Member

    class GetMemberInfo:
        data: member
        http_code: int
        trace_id: str
        result: bool

    return GetMemberInfo


def delete_member():
    temp = _EmptyReturnTemplate

    class DeleteMember:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return DeleteMember


class _Role:
    id: str
    name: str
    color: int
    hoist: int
    number: int
    member_limit: int


def get_guild_roles():
    role_ = _Role

    class GetGuildRoles:
        class data:
            guild_id: str
            roles: List[role_]
            role_num_limit: str
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return GetGuildRoles


def create_role():
    role_ = _Role

    class CreateRole:
        class data:
            role_id: str
            role: role_
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return CreateRole


def patch_role():
    role_ = _Role

    class PatchRole:
        class data:
            guild_id: str
            role_id: str
            role: role_
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return PatchRole


def delete_role():
    temp = _EmptyReturnTemplate

    class DeleteRole:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return DeleteRole


def role_members():
    class RoleMembers:
        class data:
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return RoleMembers


class _Permission_Mem:
    channel_id: str
    user_id: str
    permissions: str
    code: int
    message: str


def get_channel_member_permission():
    permission = _Permission_Mem

    class GetChannelMemberPermission:
        data: permission
        http_code: int
        trace_id: str
        result: bool

    return GetChannelMemberPermission


def put_channel_mr_permission():
    temp = _EmptyReturnTemplate

    class PutChannelMRPermission:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return PutChannelMRPermission


class _Permission_Role:
    channel_id: str
    role_id: str
    permissions: str
    code: int
    message: str


def get_channel_role_permission():
    permission = _Permission_Role

    class GetChannelRolePermission:
        data: permission
        http_code: int
        trace_id: str
        result: bool

    return GetChannelRolePermission


class _Message:
    class __MsgAttachments:
        content_type: str
        filename: str
        height: int
        width: int
        id: str
        size: int
        url: str

    class message_reference:
        message_id: str

    class author:
        bot: bool
        id: str
        username: str

    class member:
        joined_at: str
        roles: List[str]

    channel_id: str
    guild_id: str
    content: str
    id: str
    attachments = List[__MsgAttachments]
    timestamp: str
    code: int
    message: str


def get_message_info():
    message_ = _Message

    class GetMessageInfo:
        class data:
            message: message_

        http_code: int
        trace_id: str
        result: bool

    return GetMessageInfo


def send_msg():
    class SendMsg:
        class data:
            id: str
            channel_id: str
            guild_id: str
            content: str
            timestamp: str
            tts: bool
            mention_everyone: bool
            author: dict
            pinned: bool
            type: int
            flags: int
            seq_in_channel: str
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return SendMsg


def delete_msg():
    temp = _EmptyReturnTemplate

    class DeleteMsg:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return DeleteMsg


def get_guild_setting():
    class GetGuildSetting:
        class data:
            disable_create_dm: bool
            disable_push_msg: bool
            channel_ids: List[str]
            channel_push_max_num: int
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return GetGuildSetting


def create_dm_guild():
    class CreateDmGuild:
        class data:
            guild_id: str
            channel_id: str
            create_time: str
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return CreateDmGuild


def mute_member():
    temp = _EmptyReturnTemplate

    class MuteMember:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return MuteMember


def mute_members():
    class MuteMembers:
        class data:
            user_ids: List[str]
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return MuteMembers


def create_announce():
    class RecommendChannels:
        channel_id: str
        introduce: str

    class CreateAnnounce:
        class data:
            guild_id: str
            channel_id: str
            message_id: str
            announces_type: int
            recommend_channels: List[RecommendChannels]
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return CreateAnnounce


def delete_announce():
    temp = _EmptyReturnTemplate

    class DeleteAnnounce:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return DeleteAnnounce


def pinmsg():
    class PinMsg:
        class data:
            guild_id: str
            channel_id: str
            message_ids: List[str]
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return PinMsg


def delete_pinmsg():
    temp = _EmptyReturnTemplate

    class DeletePinMsg:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return DeletePinMsg


class __Schedule:
    class creator:
        class user:
            id: str
            username: str
            bot: bool

        nick: str
        joined_at: str

    id: str
    name: str
    description: str
    start_timestamp: str
    end_timestamp: str
    jump_channel_id: str
    remind_type: str
    code: int
    message: str


def get_schedules():
    schedule = __Schedule

    class GetSchedules:
        data: List[schedule]
        http_code: int
        trace_id: str
        result: bool

    return GetSchedules


def schedule_info():
    schedule = __Schedule

    class ScheduleInfo:
        data: schedule
        http_code: int
        trace_id: str
        result: bool

    return ScheduleInfo


def delete_schedule():
    temp = _EmptyReturnTemplate

    class DeleteSchedule:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return DeleteSchedule


def reactions():
    temp = _EmptyReturnTemplate

    class Reactions:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return Reactions


def get_reaction_users():
    class Users:
        id: str
        username: str
        avatar: str
        code: int
        message: str

    class GetReactionUsers:
        data: List[Users]
        http_code: List[int]
        trace_id: List[str]
        result: List[bool]

    return GetReactionUsers


def audio():
    temp = _EmptyReturnTemplate

    class Audio:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return Audio


class __Threads:
    class thread_info:
        class content:
            class __ForumsContent:
                class __ForumsSubContent:
                    class text:  # type = 1
                        text: str

                    class image:  # type = 2
                        class plat_image:
                            url: str
                            width: int
                            height: int
                            image_id: str

                    class video:  # type = 3
                        class plat_video:
                            class cover:
                                url: str
                                width: int
                                height: int

                            url: str
                            width: int
                            height: int
                            video_id: str

                    class url:  # type = 4
                        url: str
                        desc: str

                    type = int

                elems: List[__ForumsSubContent]
                props: object

            paragraphs: List[__ForumsContent]

        thread_id: str
        title: str
        date_time: str

    guild_id: str
    channel_id: str
    author_id: str
    code: int
    message: str


def get_threads():
    thread = __Threads

    class GetThreads:
        data: List[thread]
        http_code: List[int]
        trace_id: List[str]
        result: List[bool]

    return GetThreads


def get_thread_info():
    thread = __Threads

    class GetThreadInfo:
        data: thread
        http_code: int
        trace_id: str
        result: bool

    return GetThreadInfo


def create_thread():
    class CreateThread:
        class data:
            task_id: str
            create_time: str
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return CreateThread


def delete_thread():
    temp = _EmptyReturnTemplate

    class DeleteThread:
        data: temp
        http_code: int
        trace_id: str
        result: bool

    return DeleteThread


def get_guild_permissions():
    class Api:
        api: str
        path: str
        method: str
        desc: str
        auth_status: int

    class GetGuildPermissions:
        class data:
            apis = List[Api]
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return GetGuildPermissions


def create_permission_demand():
    class CreatePermissionDemand:
        class data:
            class api_identify:
                path: str
                method: str

            guild_id: str
            channel_id: str
            title: str
            desc: str
            code: int
            message: str

        http_code: int
        trace_id: str
        result: bool

    return CreatePermissionDemand



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
from inspect import stack
from json import loads, dumps
from json.decoder import JSONDecodeError
from aiohttp import ClientSession, WSMsgType, WSServerHandshakeError
from typing import Any, Callable
from asyncio import get_event_loop, all_tasks, sleep
from time import sleep as t_sleep
from re import split as re_split
from ._utils import objectize, treat_msg, exception_handler, exception_processor
from concurrent.futures import ThreadPoolExecutor


def __getattr__(identifier: str) -> object:
    if re_split(r'[/\\]', stack()[1].filename)[-1] not in ('qg_bot.py', '<frozen importlib._bootstrap>'):
        raise AssertionError("此为SDK内部使用文件，无法使用，注册机器人请使用from qg_bot.py import BOT")

    return globals()[identifier.__path__]


class BotWs:
    def __init__(self, session, ssl, logger, total_shard: int, shard_no: int, url: str, bot_id: str, bot_token: str,
                 bot_url: str, on_msg_function: Callable[[Any], Any], on_dm_function: Callable[[Any], Any],
                 on_delete_function: Callable[[Any], Any], is_filter_self: bool,
                 on_guild_event_function: Callable[[Any], Any], on_channel_event_function: Callable[[Any], Any],
                 on_guild_member_function: Callable[[Any], Any], on_reaction_function: Callable[[Any], Any],
                 on_interaction_function: Callable[[Any], Any], on_audit_function: Callable[[Any], Any],
                 on_forum_function: Callable[[Any], Any], on_audio_function: Callable[[Any], Any],
                 intents: int, msg_treat: bool, dm_treat: bool, on_start_function: Callable[[], Any], is_async: bool,
                 max_workers: int):
        """
        此为SDK内部使用类，注册机器人请使用from qg_botsdk.qg_bot import BOT

        .. seealso::
            更多教程和相关资讯可参阅：
            https://qg-botsdk.readthedocs.io/zh_CN/latest/快速入门.html
        """
        if re_split(r'[/\\]', stack()[1].filename)[-1] not in ('qg_bot.py', '<frozen importlib._bootstrap>'):
            raise AssertionError("此为SDK内部使用类，无法使用，注册机器人请使用from qg_botsdk.qg_bot import BOT")
        self.session = session
        self.__ssl = ssl
        self.logger = logger
        self.total_shard = total_shard
        self.shard_no = shard_no
        self.url = url
        self.bot_id = bot_id
        self.bot_token = bot_token
        self.bot_url = bot_url
        self.on_start_function = on_start_function
        self.on_msg_function = on_msg_function
        self.on_dm_function = on_dm_function
        self.on_delete_function = on_delete_function
        self.is_filter_self = is_filter_self
        self.on_forum_function = on_forum_function
        if not intents:
            self.logger.error('[错误] 未订阅任何事件（未订阅任何事件将导致op=9的报错，请最少订阅一个事件）')
            exit()
        self.intents = intents
        self.msg_treat = msg_treat
        self.dm_treat = dm_treat
        self.robot = None
        self.heartbeat_time = 0
        self.loop = get_event_loop()
        self.s = None
        self.reconnect_times = 0
        self.re_connect = False
        self.running = True
        self.session_id = 0
        self.flag = False
        self.heartbeat = None
        self.op9_flag = False
        self.is_async = is_async
        self.events = {"GUILD_CREATE": on_guild_event_function, "GUILD_UPDATE": on_guild_event_function,
                       "GUILD_DELETE": on_guild_event_function, "CHANNEL_CREATE": on_channel_event_function,
                       "CHANNEL_UPDATE": on_channel_event_function, "CHANNEL_DELETE": on_channel_event_function,
                       "GUILD_MEMBER_ADD": on_guild_member_function, "GUILD_MEMBER_UPDATE": on_guild_member_function,
                       "GUILD_MEMBER_REMOVE": on_guild_member_function, "MESSAGE_REACTION_ADD": on_reaction_function,
                       "MESSAGE_REACTION_REMOVE": on_reaction_function, "INTERACTION_CREATE": on_interaction_function,
                       "MESSAGE_AUDIT_PASS": on_audit_function, "MESSAGE_AUDIT_REJECT": on_audit_function,
                       "AUDIO_START": on_audio_function, "AUDIO_FINISH": on_audio_function,
                       "AUDIO_ON_MIC": on_audio_function, "AUDIO_OFF_MIC": on_audio_function}
        self.threads = ThreadPoolExecutor(max_workers) if not self.is_async else None

    async def send_connect(self):
        connect_paras = {
            "op": 2,
            "d": {
                "token": f"Bot {self.bot_id}.{self.bot_token}",
                "intents": self.intents,
                "shard": [self.shard_no, self.total_shard]
            }
        }
        await self.ws_send(dumps(connect_paras))

    async def send_reconnect(self):
        reconnect_paras = {
            "op": 6,
            "d": {
                "token": f"Bot {self.bot_id}.{self.bot_token}",
                "session_id": self.session_id,
                "seq": self.s
            }
        }
        await self.ws_send(dumps(reconnect_paras))

    async def ws_send(self, msg):
        if not self.ws.closed:
            await self.ws.send_str(msg)

    async def heart(self):
        heart_json = {"op": 1, "d": None}
        while True:
            await sleep(self.heartbeat_time)
            if not self.ws.closed:
                heart_json['d'] = self.s
                await self.ws.send_str(dumps(heart_json))

    def start_heartbeat(self):
        tasks = [task.get_name() for task in all_tasks()]
        if 'heartbeat_task' not in tasks:
            self.heartbeat = self.loop.create_task(self.heart())
            self.heartbeat.set_name('heartbeat_task')

    def get_robot_info(self, retry=False):
        robot_info = self.session.get(f'{self.bot_url}/users/@me').json()
        if 'id' not in robot_info:
            if not retry:
                return self.get_robot_info(retry)
            else:
                self.logger.error('获取机器人信息失败，机器人启动失败，程序将退出运行')
                exit()
        return objectize(robot_info)

    @exception_processor
    async def async_start_task(self, func, *args):
        await func(*args)

    @exception_processor
    def start_task(self, func, *args):
        func(*args)

    async def distribute(self, function, data):
        data["d"]["t"] = data["t"]
        data["d"]["event_id"] = data["id"]
        if function is not None:
            if not self.is_async:
                self.threads.submit(self.start_task(function, objectize(data["d"])))
            else:
                self.loop.create_task(self.async_start_task(function, objectize(data["d"])))

    @exception_processor
    async def data_process(self, data):
        t = data["t"]
        if t in ("AT_MESSAGE_CREATE", 'MESSAGE_CREATE'):
            if self.msg_treat:
                raw_msg = '' if 'content' not in data["d"] else data["d"]["content"].strip()
                at = f'<@!{self.robot.id}>'
                treated_msg = raw_msg if raw_msg.find(at) else raw_msg.replace(at, '', 1)
                data["d"]["treated_msg"] = treat_msg(treated_msg.strip())
            await self.distribute(self.on_msg_function, data)
        elif t in ("MESSAGE_DELETE", "PUBLIC_MESSAGE_DELETE", "DIRECT_MESSAGE_DELETE"):
            if self.is_filter_self:
                target = data['d']['message']['author']['id']
                op_user = data['d']['op_user']['id']
                if op_user == target:
                    return
            await self.distribute(self.on_delete_function, data)
        elif t == "DIRECT_MESSAGE_CREATE":
            if self.dm_treat:
                raw_msg = '' if 'content' not in data["d"] else data["d"]["content"].strip()
                data["d"]["treated_msg"] = treat_msg(raw_msg)
            await self.distribute(self.on_dm_function, data)
        elif t in ("FORUM_THREAD_CREATE", "FORUM_THREAD_UPDATE", "FORUM_THREAD_DELETE", "FORUM_POST_CREATE",
                   "FORUM_POST_DELETE", "FORUM_REPLY_CREATE", "FORUM_REPLY_DELETE", "FORUM_PUBLISH_AUDIT_RESULT"):
            for items in ("content", "title"):
                try:
                    data["d"]["thread_info"][items] = loads(data["d"]["thread_info"][items])
                except JSONDecodeError:
                    pass
            await self.distribute(self.on_forum_function, data)
        else:
            if t in self.events:
                func = self.events[t]
                if func:
                    await self.distribute(func, data)
            else:
                self.logger.warning(f'unknown event type: [{t}]')

    async def main(self, msg):
        data = loads(msg)
        op = data.get('op', None)
        if "s" in data:
            self.s = data["s"]
        if op == 11:
            self.logger.debug('心跳发送成功')
        elif op == 9:
            if not self.op9_flag:
                self.op9_flag = True
                if not self.re_connect:
                    await self.send_connect()
                else:
                    await self.send_reconnect()
                return
            else:
                self.logger.error('[错误] 参数出错（一般此报错为传递了无权限的事件订阅，请检查是否有权限订阅相关事件）')
                exit()
        elif op == 10:
            self.heartbeat_time = float(int(data["d"]["heartbeat_interval"]) * 0.001)
            if not self.re_connect:
                await self.send_connect()
            else:
                await self.send_reconnect()
        elif op == 0:
            if data["t"] == "READY":
                self.session_id = data["d"]["session_id"]
                self.reconnect_times = 0
                self.start_heartbeat()
                self.logger.info('连接成功，机器人开始运行')
                if not self.flag:
                    self.flag = True
                    self.robot = self.get_robot_info()
                    self.logger.info(f'机器人频道用户ID：{self.robot.id}')
                    if self.on_start_function is not None:
                        if self.is_async:
                            self.loop.create_task(self.on_start_function())
                        else:
                            self.on_start_function()
            elif data["t"] == "RESUMED":
                self.reconnect_times = 0
                self.start_heartbeat()
                self.logger.info('重连成功，机器人继续运行')
            else:
                await self.data_process(data)

    async def connect(self):
        self.reconnect_times += 1
        try:
            async with ClientSession() as ws_session:
                async with ws_session.ws_connect(self.url, ssl=self.__ssl) as self.ws:
                    while not self.ws.closed:
                        message = await self.ws.receive()
                        if message.type == WSMsgType.TEXT:
                            if not self.running:
                                if self.heartbeat is not None and not self.heartbeat.cancelled():
                                    self.heartbeat.cancel()
                                await self.ws.close()
                                self.logger.info('WS进程已结束')
                                return
                            await self.main(message.data)
                        elif message.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.ERROR):
                            if self.running:
                                self.re_connect = True
                                if self.heartbeat is not None and not self.heartbeat.cancelled():
                                    self.heartbeat.cancel()
                                self.logger.warning('BOT_WS链接已断开，正在尝试重连……')
                                return
        except Exception as e:
            self.logger.warning('BOT_WS链接已断开，正在尝试重连……')
            if self.heartbeat is not None and not self.heartbeat.cancelled():
                self.heartbeat.cancel()
            self.logger.error(e)
            self.logger.debug(exception_handler(e))
            return

    def starter(self):
        self.loop.run_until_complete(self.connect())
        while self.running:
            self.re_connect = False if self.reconnect_times >= 20 else True
            try:
                self.loop.run_until_complete(self.connect())
            except WSServerHandshakeError:
                self.logger.warning('网络连线不稳定或已断开，请检查网络链接')
            t_sleep(5)
